@@ -1,10 +1,9 @@
 import os
 
 import tiktoken
+from openai import AsyncAzureOpenAI
 from typing_extensions import Self
 
-from openai import AzureOpenAI
-
 from diskurs.llm_client import BaseOaiApiLLMClient
 from diskurs.registry import register_llm
 
@@ -29,7 +28,6 @@ class AzureOpenAIClient(BaseOaiApiLLMClient):
         tokenizer = tiktoken.encoding_for_model(model)
 
         client_params = {
-            "api_key": api_key,
             "api_version": api_version,
             "azure_endpoint": azure_endpoint or os.getenv("AZURE_OPENAI_ENDPOINT"),
         }
@@ -38,7 +36,15 @@ class AzureOpenAIClient(BaseOaiApiLLMClient):
             client_params["azure_ad_token_provider"] = get_bearer_token_provider(
                 DefaultAzureCredential(), "https://cognitiveservices.azure.com/.default"
             )
+        else:
+            client_params["api_key"] = api_key
 
-        client = AzureOpenAI(**client_params)
+        client = AsyncAzureOpenAI(**client_params)
 
-        return cls(client=client, model=model, tokenizer=tokenizer, max_tokens=max_tokens, max_repeat=3)
+        return cls(
+            client=client,
+            model=model,
+            tokenizer=tokenizer,
+            max_tokens=max_tokens,
+            max_repeat=3,
+        )
